import sys
import typing as t

from click import Option
from uvicorn import server, supervisors
from uvicorn.config import Config

from unfazed.command import BaseCommand


class Command(BaseCommand):
    help_text = """

    Run the development server

    Usage:

    >>> python manage.py runserver \n
    >>> python manage.py runserver --host 0.0.0.0 \n
    >>> python manage.py runserver --port 8000 \n
    """

    @t.override
    def add_arguments(self) -> t.List[Option]:
        return [
            Option(
                ["--host"],
                type=str,
                help="The host to listen on",
                default="127.0.0.1",
                show_default=True,
            ),
            Option(
                ["--port"],
                type=int,
                help="The port to listen on",
                default=9527,
                show_default=True,
            ),
            Option(
                ["--reload"],
                is_flag=True,
                help="Enable auto-reload",
                default=True,
                show_default=True,
            ),
            Option(
                ["--workers"],
                type=int,
                help="Number of worker processes",
                default=1,
                show_default=True,
            ),
            Option(
                ["--log-level"],
                type=str,
                help="Log level",
                default="info",
                show_default=True,
            ),
        ]

    async def handle(self, **options: t.Any) -> None:
        host = options["host"]
        port = options["port"]
        _reload = options["reload"]
        workers = options["workers"]
        log_level = options["log_level"]

        config = Config(
            self.unfazed,
            host=host,
            port=int(port),
            reload=_reload,
            workers=workers,
            log_level=log_level,
        )
        runner = server.Server(config)

        sock = config.bind_socket()
        if config.should_reload:
            supervisors.ChangeReload(config, target=runner.run, sockets=[sock]).run()

        elif config.workers > 1:
            supervisors.Multiprocess(config, target=runner.run, sockets=[sock]).run()

        else:
            await runner.serve([sock])

        if not runner.started and not config.should_reload and config.workers == 1:
            sys.exit(3)



import typing as t
from contextlib import asynccontextmanager

from starlette.types import ASGIApp

from .base import BaseLifeSpan

if t.TYPE_CHECKING:
    from unfazed.core import Unfazed  # pragma: no cover


class State(t.Dict):
    pass


class LifeSpanHandler:
    def __init__(self) -> None:
        self.lifespan: t.Dict[str, BaseLifeSpan] = {}

        self.unfazed: "Unfazed" | None = None

    def register(self, name: str, lifespan: BaseLifeSpan) -> None:
        if name in self.lifespan:
            raise ValueError(f"lifespan {name} already registered")
        self.lifespan[name] = lifespan

    def get(self, name: str) -> BaseLifeSpan | None:
        return self.lifespan.get(name)

    async def on_startup(self) -> None:
        for _, lifespan in self.lifespan.items():
            await lifespan.on_startup()

    async def on_shutdown(self) -> None:
        for _, lifespan in self.lifespan.items():
            await lifespan.on_shutdown()

    @property
    def state(self) -> State:
        ret = State()
        for _, lifespan in self.lifespan.items():
            ret.update(lifespan.state)

        return ret

    def clear(self) -> None:
        self.lifespan.clear()


handler = LifeSpanHandler()


@asynccontextmanager
async def lifespan(app: ASGIApp) -> t.AsyncIterator[State]:
    await handler.on_startup()

    yield handler.state

    await handler.on_shutdown()



import typing as t
from pathlib import Path

from tortoise import Tortoise

import unfazed
from unfazed.protocol import DataBaseDriver
from unfazed.schema import AppModels, Command, Database

if t.TYPE_CHECKING:
    from unfazed.core import Unfazed  # pragma: no cover


class Driver(DataBaseDriver):
    def __init__(self, unfazed: "Unfazed", conf: Database) -> None:
        self.conf = conf
        self.unfazed = unfazed

    async def setup(self) -> None:
        # find all models in apps
        if not self.conf.apps:
            self.conf.apps = self.build_apps()

        # init tortoise
        config = self.conf.model_dump(exclude_none=True)
        config.pop("driver")
        await Tortoise.init(config=config)

        # load aerich command
        for c in self.list_aerich_command():
            self.unfazed.command_center.load_command(c)

    async def migrate(self) -> None:
        # must call after setup
        await Tortoise.generate_schemas()

    def build_apps(self) -> t.Dict[str, t.Any]:
        models_list = ["aerich.models"]  # support aerich cmd
        for _, app in self.unfazed.app_center:
            if app.has_models():
                models_list.append(f"{app.name}.models")

        return {"models": AppModels(MODELS=models_list)}

    def list_aerich_command(self) -> t.List[Command]:
        ret = []
        internal_command_dir = Path(unfazed.__path__[0] + "/db/tortoise/commands")
        for command_file in internal_command_dir.glob("*.py"):
            command_name = command_file.stem

            path = f"unfazed.db.tortoise.commands.{command_name}.Command"
            command = Command(path=path, stem=command_name, label="aerich.command")

            ret.append(command)

        return ret



from unfazed.cache import caches
from unfazed.lifespan import BaseLifeSpan


class CacheClear(BaseLifeSpan):
    """
    CacheClear Lifespan class for unfazed.

    This class is used to clear all caches when the application is shutting down.

    """

    async def on_shutdown(self) -> None:
        await caches.close()



import typing as t

from pydantic import BaseModel

from unfazed import utils as u
from unfazed.const import UNFAZED_SETTINGS_MODULE
from unfazed.utils import Storage

T = t.TypeVar("T", bound=BaseModel)


class SettingsProxy(Storage[T]):
    """
    Settings Proxy class for unfazed

    settings will load settings from UNFAZED_SETTINGS_MODULE and cache it.

    Usage:

    ```python

    # your_settings.py

    from pydantic import BaseModel

    class YourSettings(BaseModel):

        YOUR_KEY: str
        YOUR_KEY2: int


    # settings.py

    YOUR_SETTINGS = {
        "CLIENT_CLASS": "your_settings.YourSettings",
        "YOUR_KEY": "value",
        "YOUR_KEY2": 1,

    }

    # your_services.py
    from unfazed.conf import settings

    your_settings: YourSettings = settings["YOUR_SETTINGS"]

    your_settings.YOUR_KEY
    your_settings.YOUR_KEY2

    ```

    """

    unfazed_settings_module = UNFAZED_SETTINGS_MODULE

    def __init__(self) -> None:
        super().__init__()
        self._settingskv: t.Dict[str, t.Any] | None = None

    @property
    def settingskv(self) -> t.Dict[str, t.Any]:
        if self._settingskv is None:
            self._settingskv = u.import_setting(self.unfazed_settings_module)
        return self._settingskv

    def guess_client_cls(self, alias: str) -> t.Type[T]:
        if alias == "UNFAZED_SETTINGS":
            client_str = "unfazed.conf.UnfazedSettings"

        else:
            # extract from settingskv or try to guess from alias
            alias_settings_kv = self.settingskv[alias]
            if "CLIENT_CLASS" in alias_settings_kv:
                client_str = alias_settings_kv["CLIENT_CLASS"]
            else:
                app = alias.lower().replace("_", ".").rsplit(".", 1)[0]
                alias_prefix = alias.rsplit("_", 1)[0]
                client_str = f"{app}.settings.{alias_prefix.capitalize()}Settings"

        return u.import_string(client_str)

    def __getitem__(self, key: str) -> T:
        if key in self.storage:
            return self.storage[key]

        if key not in self.settingskv:
            raise KeyError(
                f"Key {key} not found in {self.unfazed_settings_module} module"
            )

        client_cls = self.guess_client_cls(key)
        client = client_cls.model_validate(self.settingskv[key])
        self.storage[key] = client

        return client

    def clear(self) -> None:
        self._settingskv = None
        return super().clear()


settings: SettingsProxy = SettingsProxy()



import inspect
import typing as t
import warnings
from functools import wraps

from starlette.concurrency import run_in_threadpool

from .handler import caches


def cached(
    using: str = "default", timeout: int = 60, include: t.List[str] | None = None
) -> t.Callable:
    """
    Cache decorator for async function.

    Args:
        using (str): cache backend name.
        timeout (int): cache timeout.
        include (List[str]): cache key include list.

    Usage:
        ```python

        @cached(timeout=60)
        def get_user_info(user_id: int) -> dict:
            return {"user_id": user_id, "name": "Alice"}

        get_user_info(user_id=1)


        @cached(timeout=60, include=["user_id"])
        def get_user_info(user_id: int, name: str) -> dict:
            return {"user_id": user_id, "name": name}


        v1 = get_user_info(user_id=1, name="Alice")
        v2 = get_user_info(user_id=1, name="Bob")

        # v1 == v2

        @cached(using="redis", timeout=60)
        def get_user_info(user_id: int) -> dict:
            return {"user_id": user_id, "name": "Alice"}

        get_user_info(user_id=1)


        ```

    Attention:
        The cached decorator will ignore positional arguments, use keyword arguments instead.
    """

    def decorator(func: t.Callable) -> t.Callable:
        @wraps(func)
        async def wrapper(*args: t.Any, **kwargs: t.Any) -> t.Any:
            if args:
                warnings.warn(
                    "The cached decorator will ignore positional arguments, use keyword arguments instead.",
                    UserWarning,
                    stacklevel=2,
                )

            prefix = f"{func.__module__}:{func.__name__}"
            if include:
                suffix = ":".join(
                    [f"{k}_{str(v)}" for k, v in kwargs.items() if k in include]
                )

            else:
                suffix = ""

            key = f"{prefix}:{suffix}" if suffix else prefix

            force_update = kwargs.pop("force_update", False)

            cache = caches[using]
            value = await cache.get(key)
            if value and not force_update:
                return value
            else:
                if inspect.iscoroutinefunction(func):
                    result = await func(*args, **kwargs)
                else:
                    result = await run_in_threadpool(func, *args, **kwargs)
                await cache.set(key, result, timeout)
                return result

        return wrapper

    return decorator



import typing as t

from redis.asyncio import Redis
from redis.asyncio.connection import parse_url
from redis.asyncio.retry import Retry
from redis.backoff import ConstantBackoff
from unfazed.schema import RedisOptions


class DefaultBackend:
    """
    Full featured Redis cache backend

    due to redis not support prefix, so we need to handle it manually
    call `make_key` to add prefix to key

    Usage:

    ```python

    from unfazed.cache import caches

    default: DefaultBackend = caches["default"]

    await default.set(default.make_key("key"), "value")

    ret = await default.get(default.make_key("key"))

    ```

    """

    def __init__(self, location: str, options: t.Dict[str, t.Any] | None = None):
        kw = parse_url(location)
        if options is None:
            options = {}

        options_model = RedisOptions(**options)
        if options_model.retry:
            retry_cls = Retry(ConstantBackoff(0.5), 1)
        else:
            retry_cls = None

        self.prefix = options_model.prefix or ""

        self.client = Redis(
            host=kw.get("host", "localhost"),
            port=kw.get("port", 6379),
            db=kw.get("db", 0),
            password=kw.get("password", None),
            username=kw.get("username"),
            retry=retry_cls,
            socket_timeout=options_model.socket_timeout,
            socket_connect_timeout=options_model.socket_connect_timeout,
            socket_keepalive=options_model.socket_keepalive,
            socket_keepalive_options=options_model.socket_keepalive_options,
            decode_responses=options_model.decode_responses,
            retry_on_timeout=options_model.retry_on_timeout,
            retry_on_error=options_model.retry_on_error,
            max_connections=options_model.max_connections,
            single_connection_client=options_model.single_connection_client,
            health_check_interval=options_model.health_check_interval,
            ssl=options_model.ssl,
            ssl_keyfile=options_model.ssl_keyfile,
            ssl_certfile=options_model.ssl_certfile,
            ssl_cert_reqs=options_model.ssl_cert_reqs,
            ssl_ca_certs=options_model.ssl_ca_certs,
            ssl_ca_data=options_model.ssl_ca_data,
            ssl_check_hostname=options_model.ssl_check_hostname,
            ssl_min_version=options_model.ssl_min_version,
            ssl_ciphers=options_model.ssl_ciphers,
        )

    async def __aenter__(self) -> t.Self:
        return self

    async def __aexit__(self, *args: t.Any, **kw: t.Any) -> None:
        await self.client.aclose()

    async def close(self) -> None:
        # for compatibility with cache backend protocol
        return await self.client.aclose()

    def make_key(self, key: str) -> str:
        if not self.prefix:
            return key
        return f"{self.prefix}:{key}"

    def __getattr__(self, name: str) -> t.Any:
        """
        DefaultBackend will call any method of Redis client though this method

        Get all methods of Redis client, refer:
        https://redis-py.readthedocs.io/en/stable/commands.html

        Usage:

        ```python

        default: DefaultBackend = caches["default"]

        await default.set("key", "value")
        await default.get("key")

        ```
        """
        if not hasattr(self.client, name):
            raise AttributeError(f"Attribute {name} not found in Redis client")
        return getattr(self.client, name)



import importlib
import importlib.util
import typing as t
import warnings
from abc import ABC, abstractmethod
from pathlib import Path
from types import ModuleType

from unfazed.schema import Command

if t.TYPE_CHECKING:
    from unfazed.core import Unfazed  # pragma: no cover


class BaseAppConfig(ABC):
    """
    Initialize the Config object.


    Usage:

    ```python
    # app.py

    from unfazed.app import BaseAppConfig

    class AppConfig(BaseAppConfig):
        async def ready(self):
            print("App is ready!")
    ```

    """

    def __init__(self, unfazed: "Unfazed", app_module: ModuleType) -> None:
        self.unfazed = unfazed
        self.app_module = app_module

    @property
    def app_path(self) -> Path:
        if hasattr(self.app_module.__path__, "_path"):
            ret = Path(self.app_module.__path__._path[0])
        else:
            ret = Path(self.app_module.__path__[0])
        return ret

    @property
    def name(self) -> str:
        return self.app_module.__name__

    @property
    def label(self) -> str:
        return self.name.replace(".", "_")

    def list_command(self) -> t.List[Command]:
        ret: t.List[Command] = []
        commands_path = self.app_path.joinpath("commands")
        if commands_path.exists() and commands_path.is_dir():
            command_file_list = commands_path.glob("*.py")

            for command_file in command_file_list:
                if command_file.stem.startswith("_"):
                    continue

                path = f"{self.name}.commands.{command_file.stem}.Command"
                ret.append(
                    Command(
                        path=path,
                        label=self.name,
                        stem=command_file.stem,
                    )
                )

        return ret

    def has_models(self) -> bool:
        # check if models module exists
        existed = importlib.util.find_spec(f"{self.app_module.__name__}.models")

        return bool(existed)

    def wakeup(self, filename: str) -> bool:
        """
        Wake up the app with the given label.

        Args:
            filename (str): The file of the app to wake up.

        Returns:
            bool
        """
        # check if the app has the given file
        try:
            importlib.import_module(f"{self.app_module.__name__}.{filename}")
            return True
        except ImportError:
            warnings.warn(
                f"Could not import {self.app_module.__name__}.{filename}, please check if the {filename} file exists.",
                stacklevel=2,
            )

            return False

    @classmethod
    def from_entry(cls, entry: str, unfazed: "Unfazed") -> t.Self:
        """
        Load the configuration from the given entry module.

        Returns:
            Self: An instance of the configuration class.

        Raises:
            ImportError: If the module specified in `entry` cannot be imported.
            ModuleNotFoundError: If the `app.py` module cannot be found in the specified `entry` module.
            AttributeError: If `AppConfig` is not found in the `app.py` module.
            TypeError: If `AppConfig` is not a subclass of `BaseAppConfig`.
        """

        try:
            module = importlib.import_module(entry)
        except Exception:
            raise ImportError(f"Could not import module {entry}")

        # check if module has app submodule
        ret = importlib.util.find_spec(f"{module.__name__}.app")
        if ret is None:
            raise ModuleNotFoundError(f"Could not find app.py in {entry}")

        app_module = importlib.import_module(f"{module.__name__}.app")

        # check config class
        if not hasattr(app_module, "AppConfig"):
            raise AttributeError(f"AppConfig not found in {entry}.app")

        app_cls = app_module.AppConfig

        # check if AppConfig is a subclass of BaseAppConfig
        if not issubclass(app_module.AppConfig, cls):
            raise TypeError(
                f"{entry}.app.AppConfig must be a subclass of BaseAppConfig"
            )

        return app_cls(unfazed, module)

    @abstractmethod
    async def ready(self) -> None: ...

    def __str__(self) -> str:
        return self.name



import typing as t

from starlette.datastructures import MutableHeaders
from starlette.requests import HTTPConnection
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from unfazed.conf import settings
from unfazed.contrib.session.backends.base import SessionBase
from unfazed.contrib.session.settings import SessionSettings
from unfazed.contrib.session.utils import build_cookie
from unfazed.protocol import ASGIType
from unfazed.utils import import_string


class SessionMiddleware:
    def __init__(self, app: ASGIApp) -> None:
        self.app = app

        self.setting: SessionSettings = settings["SESSION_SETTINGS"]
        self.engine_cls: t.Type[SessionBase] = import_string(self.setting.engine)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ("http", "websocket"):
            await self.app(scope, receive, send)
            return

        connection = HTTPConnection(scope)

        if self.setting.cookie_name in connection.cookies:
            session_key = connection.cookies[self.setting.cookie_name]
        else:
            session_key = None

        session_store: SessionBase = self.engine_cls(self.setting, session_key)
        await session_store.load()
        scope["session"] = session_store

        async def wrapped_send(message: Message) -> None:
            if message["type"] == ASGIType.HTTP_RESPONSE_START:
                headers = MutableHeaders(scope=message)

                # if session is created / updated / deleted
                # save the session and reset the cookie
                if session_store.modified:
                    await session_store.save()

                    # if session is empty, delete the cookie
                    if session_store:
                        header_value = build_cookie(
                            self.setting.cookie_name,
                            t.cast(str, session_store.session_key),
                            max_age=self.setting.cookie_max_age,
                            expires=session_store.get_expiry_age(),
                            path=self.setting.cookie_path,
                            domain=self.setting.cookie_domain,
                            secure=self.setting.cookie_secure,
                            httponly=self.setting.cookie_httponly,
                            samesite=self.setting.cookie_samesite,
                        )
                    else:
                        header_value = build_cookie(
                            self.setting.cookie_name,
                            "null",
                            max_age=0,
                            expires=session_store.get_expiry_age(),
                            path=self.setting.cookie_path,
                            domain=self.setting.cookie_domain,
                            secure=self.setting.cookie_secure,
                            httponly=self.setting.cookie_httponly,
                            samesite=self.setting.cookie_samesite,
                        )
                    headers.append("Set-Cookie", header_value)
            await send(message)

        await self.app(scope, receive, wrapped_send)



import typing as t

from unfazed.type import CanBeImported

from .base import BaseAppConfig

if t.TYPE_CHECKING:
    from unfazed.core import Unfazed  # pragma: no cover


class AppCenter:
    """
    APP Manager

    app will be loaded and stored in the AppCenter

    """

    def __init__(
        self, unfazed: "Unfazed", installed_apps: t.List[CanBeImported]
    ) -> None:
        self.unfazed = unfazed
        self.installed_apps = installed_apps
        self._store: t.Dict[str, BaseAppConfig] = {}

    async def setup(self) -> None:
        for app_path in self.installed_apps:
            if app_path in self._store:
                raise RuntimeError(f"App with path {app_path} is already loaded")
            temp_app = self.load_app(app_path)
            await temp_app.ready()
            self._store[temp_app.name] = temp_app

        return None

    def load_app(self, app_path: str) -> BaseAppConfig:
        app_config = BaseAppConfig.from_entry(app_path, self.unfazed)
        return app_config

    def __getitem__(self, key: str) -> BaseAppConfig:
        return self._store[key]

    def __contains__(self, key: str) -> bool:
        return key in self._store

    def __iter__(self) -> t.Iterator[t.Tuple[str, BaseAppConfig]]:
        for key, value in self._store.items():
            yield (key, value)

    @property
    def store(self) -> t.Dict[str, BaseAppConfig]:
        return self._store



import typing as t


class Storage[T]:
    """
    Generic storage class.
    Use to manage data in a dictionary-like object.

    Example:
    ```python

    class CacheHandler(Storage[CacheBackend]):
        pass

    handler = CacheHandler()
    handler["key"] = CacheBackend()

    ```
    """

    def __init__(self) -> None:
        self.storage: t.Dict[str, T] = {}

    def __getitem__(self, key: str) -> T:
        try:
            return self.storage[key]
        except KeyError:
            raise KeyError(f"{self.__class__.__name__} Error: {key} not found")

    def __setitem__(self, key: str, value: T) -> None:
        self.storage[key] = value

    def __contains__(self, key: str) -> bool:
        return key in self.storage

    def __delitem__(self, key: str) -> None:
        del self.storage[key]

    def __iter__(self) -> t.Iterator[t.Tuple[str, T]]:
        for key, value in self.storage.items():
            yield key, value

    def clear(self) -> None:
        self.storage.clear()



from unfazed.type import CacheBackend
from unfazed.utils import Storage


class CacheHandler(Storage[CacheBackend]):
    """
    Cache Handler class for unfazed.


    Usage:

    ```python

    from unfazed.cache import caches


    default = caches["default"]

    await default.set("key", "value")
    ret = await default.get("key")

    ```

    """

    async def close(self) -> None:
        for backend in self.storage.values():
            await backend.close()


caches: CacheHandler = CacheHandler()



import typing as t

from starlette.concurrency import run_in_threadpool
from starlette.datastructures import State
from starlette.routing import Router
from starlette.types import ASGIApp, Receive, Scope, Send

from unfazed import protocol as p
from unfazed.app import AppCenter
from unfazed.cache import caches
from unfazed.command import CliCommandCenter, CommandCenter
from unfazed.conf import UnfazedSettings
from unfazed.conf import settings as settings_proxy
from unfazed.db import ModelCenter
from unfazed.lifespan import BaseLifeSpan, lifespan_context, lifespan_handler
from unfazed.logging import LogCenter
from unfazed.openapi import OpenApi
from unfazed.openapi.routes import patterns
from unfazed.route import Route, parse_urlconf
from unfazed.schema import LogConfig
from unfazed.utils import import_string, unfazed_locker


class Unfazed:
    """
    Core Asgi Application implementation

    Usage:

    ```python

    import os
    from unfazed import Unfazed

    async def main():

        os.environ["UNFAZED_SETTINGS_MODULE"] = "path.to.settings"
        app = Unfazed()
        await app.setup()

    ```

    Step to setup Unfazed:

    1. setup settings
        unfazed loads settings from os.environ["UNFAZED_SETTINGS_MODULE"], see `unfazed.conf.UnfazedSettings`

    2. setup logging
        unfazed setup logging from settings.LOGGING, unfazed has default logging config
        and default config will merge with settings.LOGGING

    3. setup cache
        unfazed setup cache from settings.CACHE
        unfazed provide two cache backend: Memory Backend, Redis Backend
        if settings.CACHE is empty, unfazed will skip

    4. setup app center
        unfazed setup app center from settings.INSTALLED_APPS
        unfazed loop through all apps and call app.setup()

    5. setup model center
        unfazed setup model center from settings.DATABASE
        now unfazed only support tortoise orm

    6. setup routes
        unfazed use settings.ROOT_URLCONF to setup routes as the entry point
        then loop through all apps and extract routes from app.routes

    7. setup middleware
        unfazed setup middleware from settings.MIDDLEWARE

    8. setup command center
        unfazed loop through all apps and extract commands from app.commands

    9. setup lifespan
        unfazed setup lifespan from settings.LIFESPAN

    10. setup openapi
        unfazed setup openapi from settings.OPENAPI

    """

    def __init__(
        self,
        *,
        routes: t.List[Route] | None = None,
        middlewares: t.List[t.Type[p.MiddleWare]] | None = None,
        settings: UnfazedSettings | None = None,
    ) -> None:
        self._ready = False
        self._loading = False

        self._app_center: AppCenter | None = None
        self._command_center: CommandCenter | None = None
        self._model_center: ModelCenter | None = None

        # convinient for building test
        if settings:
            settings_proxy["UNFAZED_SETTINGS"] = settings
        self._settings = settings

        self.state = State()
        self.router = Router(routes)
        self.user_middleware = middlewares or []
        self.middleware_stack: ASGIApp | None = None

    @property
    def debug(self) -> bool:
        return self.settings.DEBUG

    @property
    def settings(self) -> UnfazedSettings:
        if self._settings is None:
            self._settings = settings_proxy["UNFAZED_SETTINGS"]
        return self._settings

    @property
    def ready(self) -> bool:
        return self._ready

    @property
    def app_center(self) -> AppCenter:
        if self._app_center is None:
            self._app_center = AppCenter(self, self.settings.INSTALLED_APPS)
        return self._app_center

    @property
    def command_center(self) -> CommandCenter:
        if self._command_center is None:
            self._command_center = CommandCenter(
                self,
                self.app_center,
                self.settings.PROJECT_NAME,
            )
        return self._command_center

    @property
    def cli_command_center(self) -> CliCommandCenter:
        if hasattr(self, "_cli_command_center"):
            return self._cli_command_center
        else:
            cmd = CliCommandCenter(self)
            setattr(self, "_cli_command_center", cmd)
            return cmd

    @property
    def model_center(self) -> ModelCenter:
        if self._model_center is None:
            self._model_center = ModelCenter(
                self,
                self.settings.DATABASE,
            )
        return self._model_center

    @property
    def routes(self) -> t.List[Route]:
        return self.router.routes  # type: ignore

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if not self.ready:
            await self.setup()
        scope["app"] = self
        if self.middleware_stack is None:
            self.middleware_stack = self.build_middleware_stack()
        await self.middleware_stack(scope, receive, send)

    async def migrate(self) -> None:
        await self.model_center.migrate()

    def setup_routes(self) -> None:
        if not self.settings.ROOT_URLCONF:
            return
        # add routes from settings.ROOT_URLCONF
        for route in parse_urlconf(self.settings.ROOT_URLCONF, self.app_center):
            self.router.routes.append(route)

    def setup_middleware(self) -> None:
        if not self.settings.MIDDLEWARE:
            return
        for middleware in self.settings.MIDDLEWARE:
            cls = import_string(middleware)
            self.user_middleware.append(cls)

    def setup_cache(self) -> None:
        cache = self.settings.CACHE
        if not cache:
            return

        for alias, conf in cache.items():
            backend_cls = import_string(conf.BACKEND)
            location = conf.LOCATION
            options = conf.OPTIONS
            cache_backend = backend_cls(location, options)
            caches[alias] = cache_backend

    def setup_logging(self) -> None:
        if not self.settings.LOGGING:
            config = {}

        else:
            config = LogConfig.model_validate(self.settings.LOGGING).model_dump(
                exclude_none=True, by_alias=True
            )
        log_center = LogCenter(self, config)
        log_center.setup()

    def setup_lifespan(self) -> None:
        lifespan_handler.unfazed = self

        lifespan_list = self.settings.LIFESPAN or []

        for name in lifespan_list:
            cls = import_string(name)
            instance = cls(self)
            if not isinstance(instance, BaseLifeSpan):
                raise ValueError(f"{name} is not a valid lifespan")
            lifespan_handler.register(name, instance)

        self.router.lifespan_context = lifespan_context

    def setup_openapi(self) -> None:
        if not self.settings.OPENAPI:
            return

        OpenApi.create_schema(
            t.cast(t.List[Route], self.router.routes),
            self.settings.OPENAPI,
        )

        pattern: Route
        for pattern in patterns:  # type: ignore
            self.router.routes.append(pattern)

    def build_middleware_stack(self) -> ASGIApp:
        middleware = self.user_middleware
        app = self.router
        for cls in reversed(middleware):
            app = cls(app)  # type: ignore
        return app

    @unfazed_locker
    async def setup(self) -> None:
        """
        Setup Order is very important, refer above docstring for more info

        """
        self.setup_logging()
        self.setup_cache()
        await self.app_center.setup()
        await self.model_center.setup()
        self.setup_routes()
        self.setup_middleware()
        await self.command_center.setup()
        self.setup_lifespan()
        self.setup_openapi()

    @unfazed_locker
    async def setup_cli(self) -> None:
        self.cli_command_center.setup()

    async def execute_command_from_argv(self) -> None:
        await run_in_threadpool(self.command_center.main)

    async def execute_command_from_cli(self) -> None:
        await run_in_threadpool(self.cli_command_center.main)



import typing as t

from unfazed.protocol import DataBaseDriver
from unfazed.schema import Database
from unfazed.utils import import_string

if t.TYPE_CHECKING:
    from unfazed.core import Unfazed  # pragma: no cover


class ModelCenter:
    """
    ModelCenter is a registry for orm pkg.

    """

    def __init__(self, unfazed: "Unfazed", conf: Database | None) -> None:
        self.unfazed = unfazed
        self.conf = conf
        self.driver: DataBaseDriver | None = None

    async def setup(self) -> None:
        if not self.conf:
            return
        driver_cls = import_string(self.conf.driver)
        driver: DataBaseDriver = driver_cls(self.unfazed, self.conf)
        await driver.setup()

        self.driver = driver

    async def migrate(self) -> None:
        """
        migrate all models to database
        better using this in test case
        """
        if self.driver is None:
            raise ValueError("driver not setup")  # type: ignore
        await self.driver.migrate()



import typing as t
from logging import LogRecord

import orjson as json

JSON_STR = t.Annotated[str, "json_str"]


class UnfazedJsonFormatter:
    def __init__(self, fields: t.List[str]) -> None:
        self.fields = fields

    def format(self, record: LogRecord) -> JSON_STR:
        body = {
            field: getattr(record, field)
            for field in self.fields
            if hasattr(record, field)
        }
        return json.dumps(body).decode("utf-8")



import os
import time
import typing as t
from logging import LogRecord
from logging.handlers import BaseRotatingHandler


class UnfazedRotatingFileHandler(BaseRotatingHandler):
    """
    Process safe logging handler

    this handler will create unique name for each process log file


    """

    def __init__(
        self,
        filename: str,
        mode: str = "a",
        maxBytes: int = 0,
        encoding: str | None = None,
        delay: bool = False,
    ) -> None:
        self.raw_filename = filename
        self.validate_filename(filename)
        filename = self.create_process_safe_name(filename)
        super().__init__(filename, mode, encoding, delay)
        self.maxBytes = maxBytes

    def validate_filename(self, filename: str) -> None:
        if not filename:
            raise ValueError("filename cannot be empty")

        if not filename.endswith(".log"):
            raise ValueError("filename must end with .log")

    def create_process_safe_name(self, filename: str) -> str:
        # create a process safe name with pid
        name, suffix = filename.rsplit(".", 1)
        return f"{name}_pid{os.getpid()}_ts{int(time.time())}.{suffix}"

    @t.override
    def namer(self, filename: str) -> str:  # type: ignore
        name, suffix = filename.rsplit(".", 1)
        return f"{name}_archived.{suffix}"

    def shouldRollover(self, record: LogRecord) -> bool:
        """
        Determine if rollover should occur.

        Basically, see if the supplied record would cause the file to exceed
        the size limit we have.
        """
        if os.path.exists(self.baseFilename) and not os.path.isfile(self.baseFilename):
            return False
        if self.stream is None:  # delay was set...
            self.stream = self._open()  # pragma: no cover  # TODO: add test
        if self.maxBytes > 0:  # are we rolling over?
            msg = "%s\n" % self.format(record)
            self.stream.seek(0, 2)  # due to non-posix-compliant Windows feature
            if self.stream.tell() + len(msg) >= self.maxBytes:
                return True
        return False

    def doRollover(self) -> None:
        """
        Do a rollover, as described in __init__().
        """
        if self.stream:
            self.stream.close()
            self.stream = None  # type: ignore

        archived_filename = self.rotation_filename(self.baseFilename)
        if os.path.exists(archived_filename):
            os.remove(archived_filename)

        os.rename(self.baseFilename, archived_filename)

        new_filename = self.create_process_safe_name(self.raw_filename)
        self.baseFilename = new_filename

        if not self.delay:
            self.stream = self._open()



import os
import typing as t
from datetime import datetime
from functools import partial
from urllib.parse import quote
from zoneinfo import ZoneInfo

import anyio
import orjson as json
from pydantic import BaseModel
from starlette.background import BackgroundTask
from starlette.concurrency import iterate_in_threadpool
from starlette.responses import Response
from starlette.types import Receive, Scope, Send

from unfazed.protocol import ASGIType
from unfazed.type import ContentStream

T = t.TypeVar("T", bound=t.Union[t.Dict, t.List, str, bytes, BaseModel, ContentStream])


class HttpResponse[T](Response):
    media_type = "text/plain"

    def __init__(
        self,
        content: T | None = None,
        status_code: int = 200,
        headers: t.Mapping[str, str] | None = None,
        media_type: str | None = None,
        background: BackgroundTask | None = None,
    ) -> None:
        super().__init__(content, status_code, headers, media_type, background)


class PlainTextResponse(HttpResponse[str]):
    pass


class HtmlResponse(HttpResponse[str]):
    media_type = "text/html"


class JsonResponse(HttpResponse[t.Union[BaseModel, t.Dict, t.List]]):
    """
    Json response

    Usage:

    ```python

    from pydantic import BaseModel


    resp = JsonResponse({"foo": "bar"})
    resp2 = JsonResponse([1,2,3])

    class Resp(BaseModel):
        name: str

    resp3 = JsonResponse(Resp(name="unfazed"))


    ```

    """

    media_type = "application/json"

    def render(self, content: T) -> bytes:
        if isinstance(content, BaseModel):
            ret = json.dumps(content.model_dump())
        elif isinstance(content, (dict, list)):
            ret = json.dumps(content)
        else:
            raise ValueError(f"content {content!r} must be dumpable in JsonResponse")
        return ret


class RedirectResponse(HttpResponse):
    """

    Redirect Response


    Usage:

    ```python

    from unfazed.http import RedirectResponse

    resp = RedirectResponse("https://www.google.com")

    ```

    """

    def __init__(
        self,
        url: str,
        status_code: int = 302,
        headers: t.Mapping[str, str] | None = None,
        background: BackgroundTask | None = None,
    ) -> None:
        super().__init__(
            content=b"", status_code=status_code, headers=headers, background=background
        )
        self.headers["location"] = quote(str(url), safe=":/%#?=@[]!$&'()*+,;")


# Copy most code from starlette.responses.StreamingResponse


class StreamingResponse(HttpResponse[ContentStream]):
    """
    StreamingResponse is a response class that can be used to stream large

    ```python

    from unfazed.http import StreamingResponse

    async def stream_large_file(request) -> StreamingResponse:

        def content():
            for chunk in large_file:
                yield chunk

        return StreamingResponse(content())

    ```

    """

    def __init__(
        self,
        content: ContentStream,
        status_code: int = 200,
        headers: t.Mapping[str, str] | None = None,
        media_type: str | None = None,
        background: BackgroundTask | None = None,
    ) -> None:
        if isinstance(content, t.AsyncIterable):
            self.body_iterator = content
        else:
            self.body_iterator = iterate_in_threadpool(content)
        self.status_code = status_code
        self.media_type = self.media_type if media_type is None else media_type
        self.background = background
        self.init_headers(headers)

    async def listen_for_disconnect(self, receive: Receive) -> None:
        while True:
            message = await receive()
            if message["type"] == ASGIType.HTTP_DISCONNECT:
                break

    async def stream_response(self, send: Send) -> None:
        await send(
            {
                "type": ASGIType.HTTP_RESPONSE_START,
                "status": self.status_code,
                "headers": self.raw_headers,
            }
        )
        async for chunk in self.body_iterator:
            if not isinstance(chunk, (bytes, memoryview)):
                chunk = chunk.encode(self.charset)
            await send(
                {"type": ASGIType.HTTP_RESPONSE_BODY, "body": chunk, "more_body": True}
            )

        await send(
            {"type": ASGIType.HTTP_RESPONSE_BODY, "body": b"", "more_body": False}
        )

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        async with anyio.create_task_group() as task_group:

            async def wrap(func: t.Callable[[], t.Awaitable[None]]) -> None:
                await func()
                task_group.cancel_scope.cancel()

            task_group.start_soon(wrap, partial(self.stream_response, send))
            await wrap(partial(self.listen_for_disconnect, receive))

        if self.background is not None:
            await self.background()


class RangeFileHandler:
    def __init__(
        self,
        path: str,
        download_name: str | None = None,
        chunk_size: int = 65536,
    ) -> None:
        if not os.path.exists(path):
            raise FileNotFoundError(f"File {path} not found")

        self.path = path
        self.stat = os.stat(path)

        self._file_name = download_name or path.split("/")[-1]
        self.chunk_size = chunk_size

        self.range_start = 0
        self.range_end = self.stat.st_size
        self._content_length = self.stat.st_size

        self.downloaded = 0

        self.file = open(self.path, "rb")

    @property
    def file_name(self) -> str:
        return self._file_name

    @property
    def content_length(self) -> int:
        return self._content_length

    @property
    def file_size(self) -> int:
        return self.stat.st_size

    @property
    def etag(self) -> str:
        modified = int(self.stat.st_mtime * 1000)
        return f'W/"{self.file_size}-{modified}"'

    @property
    def last_modified(self) -> str:
        tz = ZoneInfo("GMT")
        modified = int(self.stat.st_mtime)

        return datetime.fromtimestamp(modified, tz).strftime("%a, %d %b %Y %H:%M:%S %Z")

    @property
    def content_range(self) -> str:
        return f"bytes {self.range_start}-{self.range_end - 1}/{self.file_size}"

    def close(self) -> None:
        self.file.close()

    def set_range(self, start: int, end: int) -> None:
        self.range_start = start
        self.range_end = end
        self._content_length = self.range_end - self.range_start

        self.file.seek(self.range_start)

    def __iter__(self) -> t.Iterator[bytes]:
        return self

    def __next__(self) -> bytes:
        if self.downloaded >= self.range_end - self.range_start:
            self.close()
            raise StopIteration

        chunk_size = min(self.chunk_size, self.range_end - self.downloaded)
        data = self.file.read(chunk_size)
        self.downloaded += len(data)
        return data


def parse_request(
    handler: RangeFileHandler, headers: t.Dict[str, str]
) -> t.Tuple[int, int, int]:
    header_if_range = headers.get("If-Range", None)
    header_range = headers.get("Range", None)

    range_start: int | str
    range_end: int | str
    range_start, range_end = 0, handler.file_size

    # compare If-Range and etag if If-Range existed
    # if they are not equal, download the whole file
    continue_download = True
    status_code = 200
    if header_if_range:
        if header_if_range == handler.etag:
            status_code = 206
        else:
            continue_download = False

    if header_range and continue_download:
        # we only support "bytes =start-end"
        # if other range format, download the whole file

        try:
            range_start, range_end = header_range.split("=")[1].split("-")
        except Exception:
            range_start, range_end = 0, handler.file_size
        if range_end == "":
            range_end = handler.file_size

        if range_start == "":
            range_start = 0
        try:
            range_start = int(range_start)
            range_end = int(range_end)
            status_code = 206
        except Exception:
            range_start = 0
            range_end = handler.file_size
            status_code = 200

        if range_start > range_end:
            status_code = 416

    return range_start, range_end, status_code


class FileResponse(StreamingResponse):
    """
    FileResponse is a response class that can be used to stream large file
    also, it can handle range request

    ```python

    from unfazed.http import FileResponse

    async def stream_large_file(request) -> FileResponse:

        return FileResponse("path/to/file")

    ```

    """

    def __init__(
        self,
        path: str,
        filename: str | None = None,
        *,
        chunk_size: int = 65536,
        headers: t.Dict[str, str] | None = None,
        background: BackgroundTask | None = None,
    ) -> None:
        handler = RangeFileHandler(path, filename, chunk_size)

        headers = headers or {}
        range_start, range_end, status_code = parse_request(handler, headers)
        self.status_code = status_code
        handler.set_range(range_start, range_end)
        resp_headers = self.build_headers(handler)

        super().__init__(
            handler,
            status_code,
            resp_headers,
            background=background,
            media_type="application/octet-stream",
        )

    def build_headers(self, handler: RangeFileHandler) -> t.Dict[str, str]:
        headers = {
            "ETag": handler.etag,
            "Accept-Ranges": "bytes",
            "Last-Modified": handler.last_modified,
            "Content-Length": str(handler.content_length),
            "Content-Type": "application/octet-stream",
            "Content-Disposition": f"attachment; filename={handler.file_name}",
        }

        if self.status_code == 206:
            headers["Content-Range"] = handler.content_range

        return headers



import typing as t

import orjson as json
from starlette.requests import Request

if t.TYPE_CHECKING:
    from unfazed.contrib.auth.models import AbstractUser  # pragma: no cover
    from unfazed.contrib.session.backends.base import SessionBase  # pragma: no cover
    from unfazed.core import Unfazed  # pragma: no cover


class HttpRequest(Request):
    """

    HttpRequest for unfazed

    Usage:

    ```python

    from unfazed.http import HttpRequest, HttpResponse

    async def my_view(request: HttpRequest):

        # get the request body as json
        body = await request.json()

        # get the request meta info
        path = request.path
        scheme = request.scheme

        # get the session and user from the request
        # you must have the session and user middleware installed
        session = request.session

        # you can also access the user
        # you must have the authentication middleware installed
        user = request.user

        # you can also access the app
        unfazed = request.unfazed
        return HttpResponse(content="Hello, world")


    ```

    """

    @t.override
    async def json(self) -> t.Dict:
        """
        use orjson to parse the request body as json
        """
        if not hasattr(self, "_json"):
            body = await self.body()
            self._json = json.loads(body)
        return self._json

    @property
    def scheme(self) -> str:
        return self.url.scheme

    @property
    def path(self) -> str:
        return self.url.path

    @property
    @t.override
    def session(self) -> "SessionBase":  # type: ignore
        if "session" not in self.scope:
            raise ValueError(
                "SessionMiddleware must be installed to access request.session"
            )

        return self.scope["session"]

    @property
    @t.override
    def user(self) -> "AbstractUser":  # type: ignore
        if "user" not in self.scope:
            raise ValueError(
                "AuthenticationMiddleware must be installed to access request.user"
            )
        return self.scope.get("user", None)

    @property
    def unfazed(self) -> "Unfazed":
        return self.scope["app"]



import inspect
import typing as t

from starlette.routing import Route as StartletteRoute
from starlette.routing import compile_path

from unfazed.protocol import MiddleWare as MiddleWareProtocol
from unfazed.type import CanBeImported, HttpMethod
from unfazed.utils import import_string

from . import params as p
from . import utils as u
from .endpoint import EndPointDefinition, EndpointHandler


class Route(StartletteRoute):
    @t.override
    def __init__(
        self,
        path: str,
        endpoint: t.Callable[..., t.Any],
        *,
        methods: t.List[HttpMethod] | None = None,
        name: str | None = None,
        middlewares: t.List[CanBeImported] | None = None,
        app_label: str | None = None,
        tags: t.List[str] | None = None,
        include_in_schema: bool = True,
        summary: str | None = None,
        description: str | None = None,
        externalDocs: t.Dict | None = None,
        deprecated: bool | None = None,
        response_models: t.List[p.ResponseSpec] | None = None,
    ) -> None:
        if not path.startswith("/"):
            raise ValueError(f"route `{endpoint.__name__}` paths must start with '/'")

        if not inspect.isfunction(endpoint):
            raise ValueError(f"Endpoint `{endpoint}` must be a async function")

        self.path = path
        self.endpoint = endpoint
        self.name = u.get_endpoint_name(endpoint) if name is None else name
        self.include_in_schema = include_in_schema
        self.summary = summary
        self.description = description
        self.externalDocs = externalDocs
        self.deprecated = deprecated

        if methods is None:
            methods_set = {"GET", "HEAD"}
        else:
            methods_set = {method.upper() for method in methods}
            if "GET" in methods_set:
                methods_set.add("HEAD")
        self.methods = methods_set

        self.path_regex, self.path_format, self.param_convertors = compile_path(path)

        # load app_label as tags
        if not tags:
            if app_label:
                tags = [app_label]
        self.tags = tags or []
        self.response_models = response_models

        self.app_label = app_label
        self.endpoint_definition = EndPointDefinition(
            endpoint=endpoint,
            methods=self.methods,
            tags=tags or [],
            path_parm_names=self.param_convertors.keys(),
            response_models=response_models,
        )

        self.app = EndpointHandler(self.endpoint_definition)

        self.load_middlewares(middlewares or [])

    def load_middlewares(self, middlewares: t.List[CanBeImported]) -> None:
        for cls_string in reversed(middlewares):
            cls: t.Type[MiddleWareProtocol] = import_string(cls_string)
            self.app = cls(app=self.app)

    def update_path(self, new_path: str) -> None:
        self.path = new_path
        self.path_regex, self.path_format, self.param_convertors = compile_path(
            new_path
        )

        self.endpoint_definition = EndPointDefinition(
            endpoint=self.endpoint,
            methods=self.methods,
            tags=self.tags,
            path_parm_names=self.param_convertors.keys(),
            response_models=self.response_models,
        )

    def update_label(self, app_label: str) -> None:
        self.app_label = app_label
        if not self.tags:
            self.tags = [app_label]



import typing as t
from pathlib import Path

from click import Group

import unfazed
from unfazed.schema import Command
from unfazed.utils import import_string

from .base import BaseCommand

if t.TYPE_CHECKING:
    from unfazed.app import AppCenter  # pragma: no cover
    from unfazed.core import Unfazed  # pragma: no cover


class Base(Group):
    @t.override
    def __init__(self, unfazed: "Unfazed", name: str) -> None:
        super().__init__(name=name)
        self.unfazed = unfazed

    def load_command(self, command: Command) -> None:
        cls = import_string(command.path)

        if not issubclass(cls, BaseCommand):
            raise TypeError("Command must be a subclass of BaseCommand")

        escaped_name = command.stem.replace("_", "-")
        cmd = cls(
            self.unfazed,
            escaped_name,
            command.label,
        )

        self.add_command(cmd)


class CommandCenter(Base):
    """
    Command Manage Center

    CommandCenter will load all the commands from the unfazed internal and the app center

    """

    @t.override
    def __init__(self, unfazed: "Unfazed", app_center: "AppCenter", name: str) -> None:
        super().__init__(unfazed=unfazed, name=name)
        self.app_center = app_center

    async def setup(self) -> None:
        # load all the commands from the unfazed internal
        for command in self.list_internal_command():
            self.load_command(command)

        # load all the commands from the app center
        for _, app in self.app_center:
            for command in app.list_command():
                self.load_command(command)

    def list_internal_command(self) -> t.List[Command]:
        ret = []
        internal_command_dir = Path(unfazed.__path__[0] + "/command/internal")
        for command_file in internal_command_dir.glob("*.py"):
            command_name = command_file.stem

            # ignore the startproject command
            # its a cli command and should not be loaded
            if command_name == "startproject":
                continue

            path = f"unfazed.command.internal.{command_name}.Command"
            command = Command(
                path=path, stem=command_name, label="unfazed.command.internal"
            )

            ret.append(command)

        return ret


class CliCommandCenter(Base):
    """

    Command Manage Center for CLI

    CLiCommandCenter only load the startproject command

    """

    def __init__(self, unfazed: "Unfazed") -> None:
        super().__init__(unfazed=unfazed, name="unfazed-cli")
        self.cli_command = ["startproject"]

    def setup(self) -> None:
        # load all the commands from the unfazed internal
        for command in self.list_cli_command():
            self.load_command(command)

    def list_cli_command(self) -> t.List[Command]:
        ret = []
        internal_command_dir = Path(unfazed.__path__[0] + "/command/internal")
        for command_file in internal_command_dir.glob("*.py"):
            command_name = command_file.stem

            if command_name in self.cli_command:
                path = f"unfazed.command.internal.{command_name}.Command"
                command = Command(
                    path=path, stem=command_name, label="unfazed.command.internal"
                )

                ret.append(command)

        return ret



import asyncio

from unfazed.core import Unfazed


async def _main() -> None:
    app = Unfazed()
    await app.setup_cli()
    await app.execute_command_from_cli()


def main() -> None:
    asyncio.run(_main())



import typing as t

if t.TYPE_CHECKING:
    from unfazed.core import Unfazed  # pragma: no cover


class BaseLifeSpan:
    """
    Base LifeSpan class

    reference(what is lifespan):
        https://asgi.readthedocs.io/en/latest/specs/lifespan.html

    Uasge:
    ```python


    class MyLifeSpan(BaseLifeSpan):
        async def on_startup(self) -> None:
            # get app settings
            settings = self.unfazed.settings
            # do something
            # when app start up

            # connection.start()

            ...

        async def on_shutdown(self) -> None:
            # do something
            # when app shutdown

            # connection.close()

            ...
    ```

    """

    def __init__(self, unfazed: "Unfazed") -> None:
        self.unfazed = unfazed

    async def on_startup(self) -> None:
        return None

    async def on_shutdown(self) -> None:
        return None

    @property
    def state(self) -> t.Dict[str, t.Any]:
        return {}



import asyncio
import typing as t
from abc import ABC, abstractmethod

from click import Command as ClickCommand
from click import Option, Parameter

if t.TYPE_CHECKING:
    from unfazed.core import Unfazed  # pragma: no cover


class BaseCommand(ClickCommand, ABC):
    """
    Base class for all commands.

    Usage:
    ```python

    from unfazed.command import BaseCommand
    from click import Option

    class Command(BaseCommand):
        help_text = "This is a command"
        def add_arguments(self) -> t.List[Option]:
            return [
                Option(
                    ["--host"],
                    type=str,
                    help="The host to listen on",
                    default="")
                ]

        async def handle(self, **option: t.Any) -> None:

            host = option["host"]
            print(f"Host: {host}")

    # >>> python manage.py {command} --host 0.0.0.0

    ```

    """

    help_text = ""

    @t.override
    def __init__(
        self,
        unfazed: "Unfazed",
        name: str,
        app_label: str,
        *,
        context_settings: t.MutableMapping[str, t.Any] | None = None,
        callback: t.Callable[..., t.Any] | None = None,
        params: t.List[Parameter] | None = None,
        help: str | None = None,
        epilog: str | None = None,
        short_help: str | None = None,
        options_metavar: str | None = "[OPTIONS]",
        add_help_option: bool = True,
        no_args_is_help: bool = False,
        hidden: bool = False,
        deprecated: bool = False,
    ) -> None:
        help = help or self.help_text
        params = params or []
        params.extend(self.add_arguments())
        callback = self._callback

        self.unfazed = unfazed
        self.app_label = app_label

        super().__init__(
            name,
            context_settings,
            callback,
            params,
            help,
            epilog,
            short_help,
            options_metavar,
            add_help_option,
            no_args_is_help,
            hidden,
            deprecated,
        )

    def _callback(self, **option: t.Optional[t.Any]) -> None:
        if asyncio.iscoroutinefunction(self.handle):
            asyncio.run(self.handle(**option))
        else:
            # due to handle must be implemented as a coroutine
            # so this branch will never be reached
            # just for type checking
            raise NotImplementedError(
                "handle method must be a coroutine"
            )  # pragma: no cover

    def add_arguments(self) -> t.List[Option]:
        return []

    @abstractmethod
    async def handle(self, **option: t.Any) -> None: ...



import typing as t
from datetime import datetime

from redis.asyncio import Redis
from redis.asyncio.connection import parse_url
from redis.asyncio.retry import Retry
from redis.backoff import ConstantBackoff
from unfazed.schema import RedisOptions
from unfazed.utils import import_string


class SerializerBackend:
    """
    Limited Redis backend with serializer and compressor
    SerializerBackend can store any type of data

    only supports string commands, if you need more commands,
    please use DefaultBackend

    Usage:

    ```python

    from unfazed.cache import caches

    default: SerializerBackend = caches["default"]

    await default.set("key", {"foo": "bar"})
    await default.set("key2", 1)
    await default.set("key3", [1, 2, 3])

    ```


    """

    def __init__(self, location: str, options: t.Dict[str, t.Any] | None = None):
        kw = parse_url(location)
        if options is None:
            options = {}

        options_model = RedisOptions(**options)
        if options_model.retry:
            retry_cls = Retry(ConstantBackoff(0.5), 1)
        else:
            retry_cls = None

        if options_model.serializer:
            serializer = import_string(options_model.serializer)()
        else:
            serializer = None
        self.serializer = serializer

        if options_model.compressor:
            if not serializer:
                raise ValueError(
                    f"Serializer is required for compressor: {options_model.compressor}"
                )
            compressor = import_string(options_model.compressor)()
        else:
            compressor = None
        self.compressor = compressor

        self.prefix = options_model.prefix or ""

        if options_model.decode_responses:
            raise ValueError(
                "SerializerBackend does not support decode_responses = True"
            )

        self.client = Redis(
            host=kw.get("host", "localhost"),
            port=kw.get("port", 6379),
            db=kw.get("db", 0),
            password=kw.get("password", None),
            username=kw.get("username"),
            retry=retry_cls,
            socket_timeout=options_model.socket_timeout,
            socket_connect_timeout=options_model.socket_connect_timeout,
            socket_keepalive=options_model.socket_keepalive,
            socket_keepalive_options=options_model.socket_keepalive_options,
            decode_responses=options_model.decode_responses,
            retry_on_timeout=options_model.retry_on_timeout,
            retry_on_error=options_model.retry_on_error,
            max_connections=options_model.max_connections,
            single_connection_client=options_model.single_connection_client,
            health_check_interval=options_model.health_check_interval,
            ssl=options_model.ssl,
            ssl_keyfile=options_model.ssl_keyfile,
            ssl_certfile=options_model.ssl_certfile,
            ssl_cert_reqs=options_model.ssl_cert_reqs,
            ssl_ca_certs=options_model.ssl_ca_certs,
            ssl_ca_data=options_model.ssl_ca_data,
            ssl_check_hostname=options_model.ssl_check_hostname,
            ssl_min_version=options_model.ssl_min_version,
            ssl_ciphers=options_model.ssl_ciphers,
        )

    async def __aenter__(self) -> t.Self:
        return self

    async def __aexit__(self, *args: t.Any, **kw: t.Any) -> None:
        await self.client.aclose()

    async def close(self) -> None:
        # for compatibility with cache backend protocol
        return await self.client.aclose()

    def make_key(self, key: str) -> str:
        if not self.prefix:
            return key
        return f"{self.prefix}:{key}"

    def encode(self, value: t.Any) -> t.Union[int, float, bytes]:
        if isinstance(value, (int, float)):
            return value

        if self.serializer:
            value = self.serializer.dumps(value)

        if self.compressor:
            value = self.compressor.compress(value)

        return value

    def decode(self, value: bytes | None) -> bytes | str | int | float | None:
        if value is None:
            return value

        try:
            value_str = value.decode("utf-8")

            if value_str.isdigit():
                return int(value_str)
            try:
                return float(value_str)
            except ValueError:
                pass
        except UnicodeDecodeError:
            pass

        if self.compressor:
            value = self.compressor.decompress(value)

        if self.serializer:
            value = self.serializer.loads(value)

        return value

    # general commands
    async def flushdb(self, asynchronous: bool = False, **kw: t.Any) -> t.Any:
        await self.client.flushdb(asynchronous, **kw)

    async def exists(self, name: str) -> int:
        key = self.make_key(name)
        return await self.client.exists(key)

    async def expire(
        self,
        name: str,
        time: int,
        nx: bool = False,
        xx: bool = False,
        gt: bool = False,
        lt: bool = False,
    ) -> t.Any:
        key = self.make_key(name)
        return await self.client.expire(key, time, nx, xx, gt, lt)

    async def touch(self, *args: t.Any) -> int:
        args = [self.make_key(key) for key in args]
        return await self.client.touch(*args)

    async def ttl(self, name: str) -> int:
        key = self.make_key(name)
        return await self.client.ttl(key)

    # string commands with serializer and compressor
    # int and float are not influenced

    # ========  string commands using serializer and compressor  ========
    async def get(self, name: str) -> t.Any:
        key = self.make_key(name)
        value = await self.client.get(key)
        return self.decode(value)

    async def getdel(self, name: str) -> t.Any:
        key = self.make_key(name)
        value = await self.client.getdel(key)
        return self.decode(value)

    async def getex(
        self,
        name: str,
        *,
        ex: int | None = None,
        px: int | None = None,
        exat: int | datetime | None = None,
        pxat: int | datetime | None = None,
        persist: bool = False,
    ) -> t.Any:
        key = self.make_key(name)
        value = await self.client.getex(key, ex, px, exat, pxat, persist)
        return self.decode(value)

    async def getset(self, name: str, value: t.Any) -> t.Any:
        key = self.make_key(name)
        new_value = self.encode(value)
        value = await self.client.getset(key, new_value)
        return self.decode(value)

    async def mget(self, keys: t.List[str], *args: t.Any) -> t.List:
        keys = [self.make_key(key) for key in keys]
        values = await self.client.mget(keys, *args)
        return [self.decode(value) for value in values]

    async def mset(self, mapping: t.Dict[str, t.Any]) -> str:
        mapping = {
            self.make_key(key): self.encode(value) for key, value in mapping.items()
        }
        return await self.client.mset(mapping)

    async def msetnx(self, mapping: t.Dict[str, t.Any]) -> bool:
        mapping = {
            self.make_key(key): self.encode(value) for key, value in mapping.items()
        }
        return await self.client.msetnx(mapping)

    async def set(
        self,
        name: str,
        value: t.Any,
        ex: int | None = None,
        px: int | None = None,
        nx: bool = False,
        xx: bool = False,
        keepttl: bool = False,
        get: bool = False,
        exat: int | datetime | None = None,
        pxat: int | datetime | None = None,
    ) -> t.Any:
        key = self.make_key(name)
        value = self.encode(value)
        await self.client.set(key, value, ex, px, nx, xx, keepttl, get, exat, pxat)

    async def setex(self, name: str, time: int, value: t.Any) -> str:
        key = self.make_key(name)
        value = self.encode(value)
        return await self.client.setex(key, time, value)

    async def setnx(self, name: str, value: t.Any) -> bool:
        key = self.make_key(name)
        value = self.encode(value)
        return await self.client.setnx(key, value)

    async def psetex(self, name: str, time: int, value: t.Any) -> str:
        key = self.make_key(name)
        value = self.encode(value)
        return await self.client.psetex(key, time, value)

    # ==== int and float are not influenced ====
    async def decr(self, name: str, amount: int = 1) -> int:
        key = self.make_key(name)
        return await self.client.decr(key, amount)

    async def decrby(self, name: str, amount: int = 1) -> int:
        key = self.make_key(name)
        return await self.client.decrby(key, amount)

    async def incr(self, name: str, amount: int = 1) -> int:
        key = self.make_key(name)
        return await self.client.incr(key, amount)

    async def incrby(self, name: str, amount: int = 1) -> int:
        key = self.make_key(name)
        return await self.client.incrby(key, amount)

    async def incrbyfloat(self, name: str, amount: float = 1.0) -> float:
        key = self.make_key(name)
        return await self.client.incrbyfloat(key, amount)

    async def delete(self, *names: str) -> t.Any:
        names = [self.make_key(key) for key in names]
        return await self.client.delete(*names)
